import itertools
import logging
from bisect import bisect
from collections import defaultdict
from mip import Model, minimize, BINARY, xsum, OptimizationStatus, Var
from typing import List, Dict

from autobridge.Opt.DataflowGraph import Edge, Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Device.DeviceManager import DeviceU250
from autobridge.util import get_mip_model_silent

U250_inst = DeviceU250()

_logger = logging.getLogger('autobridge')

BEND_COUNT_LIMIT = 2
VERTICAL_BOUNDARY_CAPACITY = 5280
SLR_CROSSING_BOUNDARY_CAPACITY = 5760
NON_SLR_CROSSING_HORIZONTAL_BOUNDARY = 9440


class RoutingVertex:
  def __init__(self, slot_name):
    self.slot_name = slot_name
    self.slot = Slot(U250_inst, slot_name)
    self.edges = []
    self.neighbors = set()

  def __eq__(self, other):
    return self.slot_name == other.slot_name

  def __hash__(self):
    return hash(self.slot_name)

  def get_down_left_x(self):
    return self.slot.down_left_x

  def get_down_left_y(self):
    return self.slot.down_left_y


class RoutingEdge:
  def __init__(self, v1: RoutingVertex, v2: RoutingVertex, total_wire_capacity, usable_wire_capacity):
    self.vertices = [v1, v2]
    self.total_capacity = total_wire_capacity
    self.capacity = usable_wire_capacity
    v1.edges.append(self)
    v2.edges.append(self)
    v1.neighbors.add(v2)
    v2.neighbors.add(v1)

    v_names = [v.slot_name for v in self.vertices]
    self._name = min(v_names) + '_To_' + max(v_names) # in case the order of vertices is different

  def __hash__(self):
    return hash(self._name)

  def __eq__(self, other):
    return self._name == other._name


class RoutingPath:
  """
  a path contains the vertcies it passed through
  """

  def __init__(
      self,
      vertices: List[RoutingVertex],
      bend_count: int,
      length_limit: int,
      data_width: int,
      fifo_name: str,
      slot_to_usage: Dict[Slot, Dict[str, float]]
  ) -> None:
    """
    a fifo is an edge in the dataflow graph.
    use fifo_name to make the path unique
    """
    self.vertices = vertices
    self.bend_count = bend_count
    self.length_limit = length_limit
    self.data_width = data_width
    self.fifo_name = fifo_name
    self.slot_to_usage = slot_to_usage
    self.edges = []

    # use fifo name to make the path name unique
    self._name = self.fifo_name + '_' + '_'.join([v.slot_name for v in self.vertices])

  def __hash__(self):
    return hash(self._name)

  def __eq__(self, other):
    return self._name == other._name

  def init_edges(self):
    """
    find all edges belong to this path based on the vertices
    """
    for i in range(len(self.vertices)-1):
      curr = self.vertices[i]
      next = self.vertices[i+1]
      e_common = [fifo for fifo in curr.edges if fifo in next.edges]
      assert len(e_common) == 1
      self.edges.append(e_common[0])

  def is_bend(self, prev, curr, next) -> bool:
    """
    check if three vertices are aligned either vertically or horizontally
    """
    if prev.get_down_left_x() == curr.get_down_left_x() and \
       next.get_down_left_x() == curr.get_down_left_x():
      return False
    elif prev.get_down_left_y() == curr.get_down_left_y() and \
         next.get_down_left_y() == curr.get_down_left_y():
      return False
    else:
      return True

  def get_child_paths(self) -> List['RoutingPath']:
    """
    extend the current path for one more vertex.
    Filter out candidates if the bend count is over the limit
    """
    curr = self.vertices[-1]
    # if the path only has one vertex, set prev to curr to be compatible with bend test
    if len(self.vertices) == 1:
      prev = curr
    else:
      prev = self.vertices[-2]

    # if a path is too long, stop generating child paths
    if len(self.vertices) > self.length_limit:
      return []

    # attempt to extend one more context from the current tail
    child_paths = []
    for next in curr.neighbors:
      if next != prev: # disable u turn at the site
        new_bend_count = self.bend_count + int(self.is_bend(prev, curr, next))

        # limit on bend count
        if new_bend_count > BEND_COUNT_LIMIT:
          continue

        child_paths.append(
          RoutingPath(
            self.vertices + [next],
            new_bend_count,
            self.length_limit,
            self.data_width,
            self.fifo_name,
            self.slot_to_usage
          )
        )
    return child_paths

  def get_dest(self) -> RoutingVertex:
    """
    get the current tail vertex of the path
    """
    return self.vertices[-1]

  def print_paths(self):
    _logger.debug(f'path from {self.vertices[0].slot_name} to {self.vertices[-1].slot_name} with cost {self.get_cost()}')
    for v in self.vertices:
      _logger.debug(f' => {v.slot_name}')

  def get_length(self) -> int:
    """
    number of vertices in this path
    """
    return len(self.vertices)

  def get_slots_in_path(self) -> List[Slot]:
    """
    get all the Slot objects corresponding to the path
    """
    return [v.slot for v in self.vertices]

  def get_shortest_path_length(self):
    """
    get the manhatten distance beween the source and destination slots
    the distance is the number of slots
    """
    src = self.vertices[0]
    dst = self.vertices[-1]
    dist_x = abs(src.get_down_left_x() - dst.get_down_left_x() ) / 2
    dist_y = abs(src.get_down_left_y() - dst.get_down_left_y() ) / 2
    return dist_x + dist_y + 1

  def get_cost(self) -> float:
    """
    calculate a cost if this slot is selected.
    We want to pass through less utilized slots as much as possible
    use the sum of DSP and BRAM percentage of each slot * wire_length
    """
    dsp_costs = [self.slot_to_usage[v.slot]['DSP'] if v.slot in self.slot_to_usage else 0 for v in self.vertices]
    bram_costs = [self.slot_to_usage[v.slot]['BRAM'] if v.slot in self.slot_to_usage else 0  for v in self.vertices]
    lut_costs = [self.slot_to_usage[v.slot]['LUT'] if v.slot in self.slot_to_usage else 0  for v in self.vertices]

    cost = (sum(dsp_costs) + sum(bram_costs) + 0.7 * sum(lut_costs) + 10 * len(self.vertices)) * self.data_width

    return round(cost, 4)

  def get_src_slot_name(self) -> str:
    return self.vertices[0].slot_name

  def get_dst_slot_name(self) -> str:
    return self.vertices[-1].slot_name


class RoutingGraph:
  def __init__(
      self,
      routing_slot_list: List[Slot],
      slot_to_usage: Dict[Slot, Dict[str, float]],
      routing_usage_limit: int,
      detour_path_limit: int,
  ) -> None:
    """
    routing_slot_list: list of all slots available for routing
    Note that if a slot is empty, it may not appear in slot_to_usage
    slot_to_usage: resource utilization of each slot
    """
    self.slot_name_to_vertex = {}
    self.edges = []
    self.slot_to_usage = slot_to_usage  # resource usage of each slot
    self.routing_usage_limit = routing_usage_limit
    self.detour_path_limit = detour_path_limit
    self.routing_slot_list = routing_slot_list
    self.init_routing_graph()

  def init_routing_graph(self):
    """ FIXME: currently hardcoded for U250/U280 slr/half-slr slots
    """
    # create routing vertices
    for s in self.routing_slot_list:
      self.slot_name_to_vertex[s.name] = RoutingVertex(s.name)

    # create routing edges
    for s1, s2 in itertools.combinations(self.routing_slot_list, 2):
      if s1.isNeightbor(s2):
        # determine the capacity of passing wires
        if s1.getSLR() != s2.getSLR():
          assert s1.down_left_x == s2.down_left_x
          assert s1.up_right_x == s2.up_right_x
          num_cr = s1.up_right_x - s1.down_left_x + 1

          # exclude the vitis region and the laguna used by system interconnects
          if s1.up_right_x == 7:
            num_cr -= 1.5

          # each clock region has 2 laguna columns, each has 1440 wires
          total_capacity = num_cr * 2880

        else:
          total_capacity = (s1.up_right_y - s1.down_left_y + 1) * 3000

        v1 = self.slot_name_to_vertex[s1.name]
        v2 = self.slot_name_to_vertex[s2.name]
        e = RoutingEdge(v1, v2, total_capacity, total_capacity * self.routing_usage_limit)
        self.edges.append(e)

  def get_shortest_dist(self, src: RoutingVertex, dst: RoutingVertex):
    """
    number of slots in the path. Include source and sink
    assume 2x2 slots
    """
    dist_x = abs(src.get_down_left_x() - dst.get_down_left_x() ) / 2
    dist_y = abs(src.get_down_left_y() - dst.get_down_left_y() ) / 2
    return dist_x + dist_y + 1

  def get_all_paths(
      self,
      src_slot: str,
      dst_slot: str,
      data_width: int,
      fifo_name: str
  ) -> List[RoutingPath]:
    """
    run BFS to get all paths that satisfy the requirement
    The path include the source and destination
    """
    src = self.slot_name_to_vertex[src_slot]
    dst = self.slot_name_to_vertex[dst_slot]
    shortest_dist = self.get_shortest_dist(src, dst)

    init_path = RoutingPath(
      vertices = [src],
      bend_count = 0,
      length_limit = shortest_dist + self.detour_path_limit,
      data_width = data_width,
      fifo_name = fifo_name,
      slot_to_usage = self.slot_to_usage
    )

    queue = [init_path]

    paths = []
    while queue:
      curr_path = queue.pop(0)

      if curr_path.get_dest() != dst:
        queue += curr_path.get_child_paths()
      else:
        curr_path.init_edges()
        paths.append(curr_path)

    assert len(set(paths)) == len(paths)
    return paths


class ILPRouter:
  def __init__(
      self,
      fifo_list: List[Edge],
      v2s: Dict[Vertex, Slot],
      slot_to_usage: Dict[Slot, Dict[str, float]],
      routing_slot_list: List[Slot],
  ) -> None:
    """
    we need to map each fifo to a set of routing edges in the routing graph
    """
    self.fifo_list = fifo_list
    self.v2s = v2s
    self.slot_to_usage = slot_to_usage
    self.routing_slot_list = routing_slot_list

    for fifo in self.fifo_list:
      src_slot_name = self.v2s[fifo.src].getRTLModuleName()
      dst_slot_name = self.v2s[fifo.dst].getRTLModuleName()
      if src_slot_name != dst_slot_name:
        _logger.info('edge %s, width %d, from %s to %s', fifo.name, fifo.width, src_slot_name, dst_slot_name)

  def get_fifo_to_candidate_paths(
      self,
      routing_usage_limit: float,
      detour_path_limit: int
  ) -> Dict[Edge, List[RoutingPath]]:
    """
    for each edge, generate the candidate paths to select from
    """
    routing_graph = RoutingGraph(self.routing_slot_list, self.slot_to_usage, routing_usage_limit, detour_path_limit)

    fifo_to_paths = {}
    for fifo in self.fifo_list:
      src_slot_name = self.v2s[fifo.src].getRTLModuleName()
      dst_slot_name = self.v2s[fifo.dst].getRTLModuleName()
      path_candidates = routing_graph.get_all_paths(
        src_slot_name, dst_slot_name, fifo.width, fifo.name
      )

      _logger.debug(f'fifo {fifo.name} has candidate paths:')
      for path in path_candidates:
        path.print_paths()

      fifo_to_paths[fifo] = path_candidates

    return fifo_to_paths

  def get_routing_edge_to_passing_paths(
      self, fifo_to_paths: Dict[Edge, List[RoutingPath]]
  ) -> Dict[RoutingEdge, List[RoutingPath]]:
    """
    for each routing edge, get all fifo and corresponding path candidates that go through it.
    """
    routing_edge_to_paths = defaultdict(list)
    for fifo, path_list in fifo_to_paths.items():
      for path in path_list:
        for routing_edge in path.edges:
          routing_edge_to_paths[routing_edge].append(path)

    return routing_edge_to_paths

  def get_path_to_var(self, m: Model, fifo_to_paths: Dict[Edge, List[RoutingPath]]):
    """
    for each fifo, for each candidate path,
    create a variable to represent if this candidate path is selected
    """
    path_to_var = {}
    for fifo, path_list in fifo_to_paths.items():
      for path in path_list:
        assert path not in path_to_var
        # weight matching model, relax the variable to continous
        path_to_var[path] = m.add_var(var_type=BINARY, lb=0, ub=1)

    return path_to_var

  def constrain_fifo_to_one_path(
      self,
      m: Model,
      fifo_to_paths: Dict[Edge, List[RoutingPath]],
      path_to_var: Dict[RoutingPath, Var]
  ) -> None:
    """
    for each fifo, only one candidate path is selected
    """
    for paths in fifo_to_paths.values():
      m += xsum([path_to_var[path] for path in paths]) == 1

  def constrain_routing_edge_capacity(
      self,
      m: Model,
      path_to_var: Dict[RoutingPath, Var],
      routing_edge_to_paths: Dict[RoutingEdge, List[RoutingPath]]
  ) -> None:
    """
    for each routing edge, limit the paths that go through a routing edge
    """
    for routing_edge, paths in routing_edge_to_paths.items():
      m += xsum(path_to_var[path] * path.data_width for path in paths) <= routing_edge.capacity

  def add_opt_goal(
      self,
      m: Model,
      fifo_to_paths: Dict[Edge, List[RoutingPath]],
      path_to_var: Dict[RoutingPath, Var]
  ) -> None:
    """
    minimize the total length * width of all selected paths
    """
    # concatenate to get all paths
    all_paths: List[RoutingPath] = sum(fifo_to_paths.values(), [])
    m.objective = minimize(
      xsum(path_to_var[path] * path.get_cost() for path in all_paths)
    )

  def analyze_routing_results(
    self,
    fifo_to_paths,
    fifo_to_selected_path,
    routing_edge_to_selected_paths
  ) -> None:
    # log the quality of the selected path
    for fifo, paths in fifo_to_paths.items():
      if len(paths) > 1:
        all_costs = [p.get_cost() for p in paths]
        all_costs.sort()
        selected_path = fifo_to_selected_path[fifo]
        rank = bisect(all_costs, selected_path.get_cost())

        rank_info = f'{fifo.name} from {selected_path.get_src_slot_name()} to {selected_path.get_dst_slot_name()} is routed with the rank {rank} / {len(all_costs)} path. '
        overall_info = f'Path cost: {selected_path.get_cost()}. All costs: {all_costs}'
        _logger.debug(rank_info + overall_info)

    # log the utilization ratio of each boundary
    for routing_edge, paths in routing_edge_to_selected_paths.items():
      total_data_width = sum([path.data_width for path in paths])
      _logger.info(f'boundary {routing_edge._name} is passed by {total_data_width} / {routing_edge.total_capacity} = {total_data_width / routing_edge.total_capacity} ')
      for path in paths:
        _logger.debug(f'  {path.fifo_name}')

  def get_routing_results(
    self,
    fifo_to_paths: Dict[Edge, List[RoutingPath]],
    path_to_var,
    routing_edge_to_paths: Dict[RoutingEdge, List[RoutingPath]]
  ) -> Dict[str, List[Slot]]:

    # get the selected paths
    fifo_to_selected_path = {}
    for fifo, paths in fifo_to_paths.items():
      for path in paths:
        val = path_to_var[path].x
        assert abs(val - round(val)) < 0.0001
        if round(val) == 1:
          fifo_to_selected_path[fifo] = path
          _logger.debug(f'fifo {fifo.name} is routed to: ')
          path.print_paths()
          if path.get_length() > path.get_shortest_path_length():
            _logger.warning(f'{fifo.name} is not routed with the shortest paths')

          break
      assert fifo in fifo_to_selected_path

    # get which paths will pass through a boundary
    routing_edge_to_selected_paths = defaultdict(list)
    for routing_edge, paths in routing_edge_to_paths.items():
      _logger.debug(f'boundary {routing_edge._name} is passed by:')
      for path in paths:
        val = path_to_var[path].x
        if round(val) == 1:
          routing_edge_to_selected_paths[routing_edge].append(path)
          _logger.debug(f'  {path.fifo_name}')

    return fifo_to_selected_path, routing_edge_to_selected_paths

  def get_fifo_to_path_exclude_src_dst(self, fifo_to_selected_path):
    e_name_to_paths = {}
    # exclude the source and destination to feed back to the outside world
    for fifo, selected_path in fifo_to_selected_path.items():
      e_name_to_paths[fifo.name] = selected_path.get_slots_in_path()[1:-1]

    return e_name_to_paths

  def route_design(
      self,
      routing_usage_limit: float = 0.6,
      detour_path_limit: int = 4
  ) -> Dict[Edge, List[Slot]]:

    while 1:
      if routing_usage_limit > 1:
        _logger.error(f'Global routing failed')
        exit(1)

      _logger.info(f'Global routing attempt with routing usage limit {routing_usage_limit}')

      m = get_mip_model_silent()

      fifo_to_paths = self.get_fifo_to_candidate_paths(routing_usage_limit, detour_path_limit)
      path_to_var = self.get_path_to_var(m, fifo_to_paths)
      routing_edge_to_paths = self.get_routing_edge_to_passing_paths(fifo_to_paths)

      _logger.info(f'there are {len(fifo_to_paths)} dataflow edges')
      _logger.info(f'there are {len(path_to_var)} potential paths to select from')

      self.constrain_fifo_to_one_path(m, fifo_to_paths, path_to_var)

      self.constrain_routing_edge_capacity(m, path_to_var, routing_edge_to_paths)

      self.add_opt_goal(m, fifo_to_paths, path_to_var)

      status = m.optimize()

      if status == OptimizationStatus.OPTIMAL:
        _logger.warning(f'Succeeded: global routing attempt with routing usage limit {routing_usage_limit}')
        break
      else:
        _logger.warning(f'Failed: global routing attempt with routing usage limit {routing_usage_limit}')
        routing_usage_limit += 0.03

    # extract results
    fifo_to_selected_path, routing_edge_to_selected_paths = \
      self.get_routing_results(fifo_to_paths, path_to_var, routing_edge_to_paths)

    # _logger and analysis
    self.analyze_routing_results(fifo_to_paths, fifo_to_selected_path, routing_edge_to_selected_paths)

    fifo_to_slots = {fifo: path.get_slots_in_path() for fifo, path in fifo_to_selected_path.items()}
    return fifo_to_slots


if __name__ == '__main__':
  routing_graph = RoutingGraph()

  # paths = routing_graph.get_all_paths('CR_X4Y0_To_CR_X5Y1', 'CR_X4Y8_To_CR_X5Y9', 10, 'test_name')
  # print(len(paths))

  paths = routing_graph.get_all_paths('CR_X4Y2_To_CR_X5Y3', 'CR_X4Y8_To_CR_X5Y9', 10, 'test_name')
  print(len(paths))

  for path in paths:
    path.print_paths()


import ctypes
from datetime import datetime
import io
import logging
import mip
import os, sys
import pathlib
import tempfile
from contextlib import contextmanager

libc = ctypes.CDLL(None)
c_stdout = ctypes.c_void_p.in_dll(libc, 'stdout')

@contextmanager
def stdout_redirector(stream):
    # The original fd stdout points to. Usually 1 on POSIX systems.
    original_stdout_fd = sys.stdout.fileno()

    def _redirect_stdout(to_fd):
        """Redirect stdout to the given file descriptor."""
        # Flush the C-level buffer stdout
        libc.fflush(c_stdout)
        # Flush and close sys.stdout - also closes the file descriptor (fd)
        sys.stdout.close()
        # Make original_stdout_fd point to the same file as to_fd
        os.dup2(to_fd, original_stdout_fd)
        # Create a new sys.stdout that points to the redirected fd
        sys.stdout = io.TextIOWrapper(os.fdopen(original_stdout_fd, 'wb'))

    # Save a copy of the original stdout fd in saved_stdout_fd
    saved_stdout_fd = os.dup(original_stdout_fd)
    try:
        # Create a temporary file and redirect stdout to it
        tfile = tempfile.TemporaryFile(mode='w+b')
        _redirect_stdout(tfile.fileno())
        # Yield to caller, then redirect stdout back to the saved fd
        yield
        _redirect_stdout(saved_stdout_fd)
        # Copy contents of temporary file to the given stream
        tfile.flush()
        tfile.seek(0, io.SEEK_SET)
        stream.write(tfile.read())
    finally:
        tfile.close()
        os.close(saved_stdout_fd)


def get_mip_model_silent():
  f = io.BytesIO()
  with stdout_redirector(f):
    m = mip.Model()
    m.verbose = 0
  return m


def get_log_name():
  return f'autobridge-{datetime.now().strftime("%b-%d-%Y-%H:%M")}.log'


def get_file_logger():
  return logging.getLogger('autobridge')


def get_cli_logger():
  return logging.getLogger('general')


def init_logging() -> None:
  logger = get_file_logger()
  logger.parent = None
  log_file = get_log_name()
  handler = logging.FileHandler(log_file)
  handler.setFormatter(logging.Formatter('[%(levelname)s:%(module)s:%(lineno)d] %(message)s'))
  logger.addHandler(handler)
  logger.setLevel(logging.INFO)


def set_general_logger():
  general_logger = get_cli_logger()
  general_logger.parent = None
  general_logger.addHandler(logging.StreamHandler())
  general_logger.setLevel(logging.INFO)

  return general_logger


def print_start():
  with open(f'{pathlib.Path(__file__).parent.resolve()}/../VERSION') as f:
    version = f.read().strip()

  general_logger = get_cli_logger()
  general_logger.info('')
  general_logger.info('*********************************************')
  general_logger.info('***         Starting AutoBridge           ***')
  general_logger.info('*********************************************')
  general_logger.info('')
  general_logger.info(f'Version: {version}')
  general_logger.info('')
  general_logger.info('Running details logged to ' + get_log_name())
  general_logger.info('')
  general_logger.info('----------------------------------------------')
  general_logger.info('')


def print_end():
  general_logger = get_cli_logger()
  general_logger.info('*********************************************')
  general_logger.info('***          AutoBridge Finishes          ***')
  general_logger.info('*********************************************')
  general_logger.info('')



import copy
from typing import List, Dict

import autobridge.Floorplan.Utilities as util
import autobridge.Floorplan as autobridge_floorplan
from autobridge.Device.DeviceManager import DeviceManager
from autobridge.HLSParser.tapa.DataflowGraphTapa import DataflowGraphTapa
from autobridge.HLSParser.tapa.ProgramJsonManager import ProgramJsonManager
from autobridge.Opt.DataflowGraph import Edge, Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager
from autobridge.Route.global_route import ILPRouter
from autobridge.util import *
from autobridge.analyze import analyze_result, analyze_input

def annotate_floorplan(config: Dict) -> Dict:
  cli_logger = set_general_logger()

  print_start()

  init_logging()

  analyze_input(config)

  board = DeviceManager(
      board_name=get_board_num(config),
      ddr_list=get_ddr_list(config),
      is_vitis_enabled=True,
  ).getBoard()
  program_json_manager = ProgramJsonManager(
      config['edges'],
      get_vertex_section(config),
      get_area_section(config),
  )
  graph = DataflowGraphTapa(program_json_manager)
  slot_manager = SlotManager(board)

  # which modules must be assigned to the same slot
  grouping_constraints: List[List[str]] = config.get('grouping_constraints', [])

  # process optional module pre-assignment constraints
  module_floorplan = config['floorplan_pre_assignments']
  pre_assignment = {}
  for region, module_group in module_floorplan.items():
    for mod_name in module_group:
      pre_assignment[mod_name] = region

  kw_args = {}
  if 'floorplan_strategy' in config:
    kw_args['floorplan_strategy'] = config['floorplan_strategy']
  if 'floorplan_opt_priority' in config:
    kw_args['floorplan_opt_priority'] = config['floorplan_opt_priority']

  # generate floorplan
  v2s, slot_list = autobridge_floorplan.get_floorplan(
    graph,
    slot_manager,
    grouping_constraints,
    pre_assignment,
    **kw_args,
  )

  # if floorplan failed
  if v2s is None:
    cli_logger.critical('\n*** CRITICAL WARNING: AutoBridge fails to find a solution. Please refer to the log for details.\n')
    config['floorplan_status'] = 'FAILED'
    print_end()
    return config

  slot_to_usage = util.get_slot_utilization(v2s)

  # route the FIFO pipelines
  router = ILPRouter(
    list(graph.edges.values()),
    v2s,
    slot_to_usage,
    slot_list,
  )
  fifo_to_path: Dict[Edge, List[Slot]] = router.route_design()

  annotated_config = get_annotated_config(v2s, fifo_to_path, slot_to_usage, config)

  analyze_result(annotated_config)

  print_end()

  return annotated_config


def get_ddr_list(config) -> List[int]:
  """ extract which ddrs are enabled
  """
  ddr_set = set()
  for v_name, properties in config['vertices'].items():
    if properties['category'] == 'PORT_VERTEX':
      if properties['port_cat'] == 'DDR':
        ddr_set.add(properties['port_id'])
  return list(ddr_set)

def get_board_num(config) -> str:
  device_num = config['part_num']
  if device_num == 'xcu250-figd2104-2L-e':
    return 'U250'
  elif device_num == 'xcu280-fsvh2892-2L-e':
    return 'U280'
  else:
    raise NotImplementedError(f'unsupported device {device_num}')

def get_vertex_section(config) -> Dict[str, str]:
  return {v_name: properties['module'] for v_name, properties in config['vertices'].items()}

def get_area_section(config) -> Dict[str, Dict[str, int]]:
  return {properties['module']: properties['area'] for v_name, properties in config['vertices'].items()}

def get_annotated_config(
    v2s: Dict[Vertex, Slot],
    fifo_to_path: Dict[Edge, List[Slot]],
    slot_to_usage: Dict[Slot, Dict[str, float]],
    config_orig: Dict,
) -> Dict:
  config = copy.deepcopy(config_orig)
  for v, s in v2s.items():
    config['vertices'][v.name]['floorplan_region'] = s.getRTLModuleName()

  slot_list = list(set(v2s.values()))
  config['floorplan_region_pblock_tcl'] = {s.getRTLModuleName(): s.pblock_tcl for s in slot_list}

  for fifo, path in fifo_to_path.items():
    config['edges'][fifo.name]['path'] = [s.name for s in path]

  config['slot_resource_usage'] = {s.getRTLModuleName(): usage for s, usage in slot_to_usage.items()}
  config['floorplan_status'] = 'SUCCEED'

  return config



import logging
import os

from typing import Dict, List, Tuple
from autobridge.Floorplan.Partition import partition
from autobridge.Floorplan.IterativeBipartion import iterative_bipartition
from autobridge.Floorplan.Utilities import (
  print_pre_assignment,
  print_vertex_areas,
  get_eight_way_partition_slots,
  get_four_way_partition_slots,
)
from autobridge.Opt.DataflowGraph import Vertex, DataflowGraph
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir

_logger = logging.getLogger('autobridge')


def get_floorplan(
  graph: DataflowGraph,
  slot_manager: SlotManager,
  grouping_constraints_in_str: List[List[str]],
  pre_assignments_in_str: Dict[str, str],
  floorplan_strategy: str = 'HALF_SLR_LEVEL_FLOORPLANNING',
  threshold_for_iterative: int = 400,
  floorplan_opt_priority: str = 'AREA_PRIORITIZED',
) -> Tuple[Dict[Vertex, Slot], List[Slot]]:
  """
  main entrance of the floorplan part
  return (1) mapping from vertex to slot
  (2) a list of all potential slots
  Note that an empty slot will not be in (1), but will occur in (2)
  """
  # get initial v2s
  init_slot = slot_manager.getInitialSlot()
  init_v2s = {v : init_slot for v in graph.getAllVertices()}

  # get grouping constraints of Vertex
  grouping_constraints: List[List[Vertex]] = [
    [graph.getVertex(v_name) for v_name in v_name_group]
      for v_name_group in grouping_constraints_in_str
  ]

  _logger.info(f'The following modules are grouped to the same location:')
  for grouping in grouping_constraints_in_str:
    _logger.info('    ' + ', '.join(grouping))

  # get pre_assignment in Vertex
  pre_assignments = { graph.getVertex(v_name) : slot_manager.createSlot(pblock)
    for v_name, pblock in pre_assignments_in_str.items()
  }

  print_pre_assignment(pre_assignments)

  print_vertex_areas(init_v2s.keys())

  # choose floorplan method
  num_vertices = len(graph.getAllVertices())
  v2s: Dict[Vertex, Slot] = {}

  # if user specifies floorplan methods
  if floorplan_strategy == 'SLR_LEVEL_FLOORPLANNING':
    _logger.info(f'user specifies to floorplan into SLR-level slots')
    v2s = partition(
      init_v2s, slot_manager, grouping_constraints, pre_assignments, partition_method='FOUR_WAY_PARTITION', priority=floorplan_opt_priority
    )

    if v2s:
      return v2s, get_four_way_partition_slots(slot_manager)
    else:
      return None

  elif floorplan_strategy == 'QUICK_FLOORPLANNING':
    _logger.info(f'user specifies to prioritize speed')
    v2s = iterative_bipartition(init_v2s, slot_manager, grouping_constraints, pre_assignments)
    if v2s:
      return v2s
    else:
      return None

  else:
    if floorplan_strategy is not 'HALF_SLR_LEVEL_FLOORPLANNING':
      raise NotImplementedError('unrecognized floorplan strategy %s', floorplan_strategy)

  # empirically select the floorplan method
  if num_vertices < threshold_for_iterative:
    _logger.info(f'There are {num_vertices} vertices in the design, use eight way partition')

    if num_vertices > 100:
      _logger.warning('Over 100 vertices. May have a long solving time. Reduce threshold_for_iterative to skip to iterative bi-partitioning.')

    v2s = partition(
      init_v2s, slot_manager, grouping_constraints, pre_assignments, partition_method='EIGHT_WAY_PARTITION', priority=floorplan_opt_priority
    )
    if v2s:
      return v2s, get_eight_way_partition_slots(slot_manager)
    else:
      _logger.warning(f'Please check if any function in the design is too large')

  else:
    _logger.info(f'Use four-way partition because eight-way partition failed or there are too many vertices ({num_vertices})')
    v2s = partition(
      init_v2s, slot_manager, grouping_constraints, pre_assignments, partition_method='FOUR_WAY_PARTITION', priority=floorplan_opt_priority
    )
    if v2s:
      return v2s, get_four_way_partition_slots(slot_manager)

  _logger.error(f'AutoBridge fails to partition the design at the SLR level. Either the design is too large, or the functions/modules are too large.')
  return None, None



import os
from prettytable import PrettyTable
from autobridge.util import get_cli_logger
from autobridge.Floorplan.Utilities import RESOURCE_TYPES

logger = get_cli_logger()


def get_port_info(config, port_name):
  port_vertex_name = f'PORT_VERTEX_{port_name}_external_controller'
  prop = config['vertices'][port_vertex_name]
  return prop['port_cat'], prop['port_id']

def get_hbm_port_side(port_cat, port_id):
  assert port_cat == 'HBM'
  if 0 <= port_id < 16:
    return 'LEFT'
  elif 16 <= port_id < 32:
    return 'RIGHT'
  else:
    raise NotImplementedError(f'unrecognized port_id {port_id}')

def get_oppo_side(side: str):
  side_map = {
    'LEFT': 'RIGHT',
    'RIGHT': 'LEFT',
  }
  return side_map[side]

def check_port_binding(config) -> None:
  """
  check for horizontal AXI edges in SLR 0
  Ideally, each m_axi module should be together with the physical port
  If not in the floorplan results, throw a warning
  """
  if not config['part_num'].startswith('xcu280'):
    return

  wrong_binding = False
  for name, prop in config['edges'].items():
    if prop['category'] == 'AXI_EDGE':
      if 'CR_X0Y0_To_CR_X3Y3' in prop['path'] and \
         'CR_X0Y4_To_CR_X3Y7' in prop['path']:
        port_name = prop['port_name']
        port_cat, port_id = get_port_info(config, port_name)
        hbm_side = get_hbm_port_side(port_cat, port_id)
        oppo_side = get_oppo_side(hbm_side)

        source = prop['path'][0]
        dst = prop['path'][1]
        logger.critical(
          '*** CRITICAL WARNING: '
          'The top argument %s is mapped to port %s on the %s half of SLR 0. '
          'However, its adjacent module %s is floorplaned to the %s half. ',
          port_name, f'{port_cat}[{port_id}]', hbm_side, prop['consumed_by'], oppo_side)
        wrong_binding = True

  if wrong_binding:
    logger.critical('')
    logger.critical(
      '*** CRITICAL WARNING: '
      'Consider adjusting the port binding to balance resources on the two sides of SLR 0.'
    )
    logger.critical('')

def check_resource_usage(config):
  usage = config['slot_resource_usage']

  table = PrettyTable(['Slot Name'] + [f'{r} (%)' for r in RESOURCE_TYPES])
  for s_name, type_to_usage in usage.items():
    table.add_row([s_name] + [f'{round(type_to_usage[type]*100, 1)}' for type in RESOURCE_TYPES])
  logger.info(table)
  logger.info('')

  if config.get('floorplan_strategy') == 'SLR_LEVEL_FLOORPLANNING':
    logger.info('only floorplan to SLR-level slots as requested by the user')
    return

  logger.info('The device could be partitioned into %d slots:', len(usage))

  if config['part_num'].startswith('xcu250') and len(usage) <= 4 or \
     config['part_num'].startswith('xcu280') and len(usage) <= 3:
    logger.critical(
      '*** CRITICAL WARNING: '
      'Some tasks of the design may be too large and prohibits a more fine-grained floorplanning.'
    )
    logger.critical('')
    logger.critical(
      'Tips: (1) write smaller tasks; (2) make each task use less heterogeneous resources. '
      'E.g., using a lot of DSP / BRAM / URAM in the *same* task makes it harder to floorplan'
    )
    logger.critical('')


def analyze_result(config) -> None:
  check_port_binding(config)
  check_resource_usage(config)


def check_gurobi() -> None:
  if 'GUROBI_HOME' in os.environ:
    logger.info('Gurobi solver detected.')
  else:
    logger.critical('*** CRITICAL WARNING: Gurobi solver not detected. Floorplanning may take extra time.')


def analyze_input(config) -> None:
  check_gurobi()



import logging
from typing import Dict, List

from autobridge.Floorplan.Utilities import log_resource_utilization
from autobridge.Floorplan.EightWayPartition import eight_way_partition
from autobridge.Floorplan.FourWayPartition import four_way_partition
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager

_logger = logging.getLogger('autobridge')


def partition(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float = 0.65,
  max_area_limit: float = 0.85,
  min_slr_width_limit: int = 10000,
  max_slr_width_limit: int = 15000,
  max_search_time: int = 600,
  partition_method: str = 'EIGHT_WAY_PARTITION',
  priority: str = 'AREA_PRIORITIZED',
) -> Dict[Vertex, Slot]:
  _logger.info('Use partition method: %s with priority %s', partition_method, priority)

  if partition_method == 'EIGHT_WAY_PARTITION':
    partitioner = eight_way_partition
  elif  partition_method == 'FOUR_WAY_PARTITION':
    partitioner = four_way_partition
  else:
    raise NotImplementedError('unrecognized floorplan method: %s', partition_method)

  if priority == 'AREA_PRIORITIZED':
    worker = partition_area_prioritized
  elif priority == 'SLR_CROSSING_PRIORITIZED':
    worker = partition_slr_crossing_prioritized
  else:
    raise NotImplementedError('unrecognized priority: %s', priority)

  return worker(
    init_v2s,
    slot_manager,
    grouping_constraints,
    pre_assignments,
    min_area_limit,
    max_area_limit,
    min_slr_width_limit,
    max_slr_width_limit,
    max_search_time,
    partitioner,
  )


def partition_area_prioritized(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float,
  max_area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  partitioner,
) -> Dict[Vertex, Slot]:
  """
  adjust the max_usage_ratio if failed
  """
  best_v2s = {}

  # start from the largest allowed ratio and search downwards. Prune if fail.
  # for curr_max_usage in reversed(float_range(ref_usage_ratio, max_usage_ratio, 0.02)):
  lo = min_area_limit
  hi = max_area_limit
  assert lo < hi

  while (1):
    curr_area_limit = (lo + hi) / 2

    _logger.info(f'Attempt eight way partition with area limit of {curr_area_limit}')

    curr_best_v2s = _binary_search_slr_crossing_limit(
      init_v2s,
      slot_manager,
      grouping_constraints,
      pre_assignments,
      curr_area_limit,
      min_slr_width_limit,
      max_slr_width_limit,
      max_search_time,
      partitioner,
    )
    if curr_best_v2s:
      best_v2s = curr_best_v2s
      hi = curr_area_limit
    else:
      lo = curr_area_limit

    if best_v2s:
      if hi - lo < 0.03:
        break
    else:
      if hi - lo < 0.02:
        break

  if not best_v2s:
    _logger.info(f'eight way partition failed with max_usage_ratio {curr_area_limit} and slr_width_limit {max_slr_width_limit}')
    return {}

  log_resource_utilization(best_v2s)
  return best_v2s


def _binary_search_slr_crossing_limit(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  partitioner,
) -> Dict[Vertex, Slot]:

  curr_best_v2s = {}

  # binary search:
  hi = max_slr_width_limit
  lo = min_slr_width_limit
  assert lo < hi

  while (1):
    curr_slr_limit = (hi + lo) / 2
    _logger.info(f'Try slr_width_limit {curr_slr_limit}')

    v2s = partitioner(
      init_v2s,
      grouping_constraints,
      pre_assignments,
      slot_manager,
      area_limit,
      curr_slr_limit,
      max_search_time,
    )

    if v2s:
      curr_best_v2s = v2s
      curr_min_slr_limit = curr_slr_limit
      hi = curr_slr_limit
    else:
      lo = curr_slr_limit

    if curr_best_v2s:
      if hi - lo < 700:
        break
    else:
      if hi - lo < 500:
        break

  if curr_best_v2s:
    _logger.info(f'Found solution with area limit {area_limit} and slr_width_limit {curr_min_slr_limit}')

  return curr_best_v2s


def partition_slr_crossing_prioritized(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  min_area_limit: float,
  max_area_limit: float,
  min_slr_width_limit: int,
  max_slr_width_limit: int,
  max_search_time: int,
  partitioner,
) -> Dict[Vertex, Slot]:
  """
  adjust the max_usage_ratio if failed
  """
  best_v2s = {}

  # start from the largest allowed ratio and search downwards. Prune if fail.
  # for curr_max_usage in reversed(float_range(ref_usage_ratio, max_usage_ratio, 0.02)):
  lo = min_slr_width_limit
  hi = max_slr_width_limit
  assert lo < hi

  while (1):
    curr_slr_limit = (lo + hi) / 2

    _logger.info(f'Attempt eight way partition with slr_crossing_limit {curr_slr_limit}')

    curr_best_v2s = _binary_search_area_limit(
      init_v2s,
      slot_manager,
      grouping_constraints,
      pre_assignments,
      curr_slr_limit,
      min_area_limit,
      max_area_limit,
      max_search_time,
      partitioner,
    )
    if curr_best_v2s:
      best_v2s = curr_best_v2s
      hi = curr_slr_limit
    else:
      lo = curr_slr_limit

    if best_v2s:
      if hi - lo < 700:
        break
    else:
      if hi - lo < 500:
        break

  if not best_v2s:
    _logger.info(f'eight way partition failed with area limit {max_area_limit} and slr_width_limit {curr_slr_limit}')
    return {}

  log_resource_utilization(best_v2s)
  return best_v2s


def _binary_search_area_limit(
  init_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  slr_width_limit,
  min_area_limit,
  max_area_limit,
  max_search_time,
  partitioner,
) -> Dict[Vertex, Slot]:

  curr_best_v2s = {}

  # binary search:
  hi = max_area_limit
  lo = min_area_limit
  assert lo < hi

  while (1):
    curr_area_limit = (hi + lo) / 2
    _logger.info(f'Try area limit {curr_area_limit}')

    v2s = partitioner(
      init_v2s,
      grouping_constraints,
      pre_assignments,
      slot_manager,
      curr_area_limit,
      slr_width_limit,
      max_search_time,
    )

    if v2s:
      curr_best_v2s = v2s
      curr_min_usage = curr_area_limit
      hi = curr_area_limit
    else:
      lo = curr_area_limit

    if curr_best_v2s:
      if hi - lo < 0.015:
        break
    else:
      if hi - lo < 0.01:
        break

  if curr_best_v2s:
    _logger.info(f'Found solution with usage_ratio {curr_min_usage} and slr_width_limit {slr_width_limit}')

  return curr_best_v2s
